﻿using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Threading;
using SocketApp.Model;
using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Threading.Tasks;
using Windows.Networking;
using Windows.Networking.Connectivity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;

namespace SocketApp.ViewModel
{
    public class MainViewModel : ViewModelBase
    {
        private const string PortOrService = "5150";
        private StreamSocket _socket;
        private StreamSocketListener _listener;

        public MainViewModel()
        {
            ChatMessages = new ObservableCollection<ChatMessage>();
            ConnectionStatus = "Not connected.";
            ServerAddress = "pete-surface64";

            CreateNewMessage();

            PostNewMessageCommand = new RelayCommand(
                () => PostNewMessage(),
                () => CanPostNewMessage());

            ListenCommand = new RelayCommand(
                () => Listen(),
                () => CanListen());

            ConnectCommand = new RelayCommand(
                () => Connect(),
                () => CanConnect());
        }

        public ObservableCollection<ChatMessage> ChatMessages { get; set; }

        private ChatMessage _newMessage;
        public ChatMessage NewMessage
        {
            get { return _newMessage; }
            set { Set<ChatMessage>(() => NewMessage, ref _newMessage, value); }
        }

        public RelayCommand PostNewMessageCommand
        {
            get; private set;
        }

        private void CreateNewMessage()
        {
            if (NewMessage != null)
                NewMessage.PropertyChanged -= NewMessage_PropertyChanged;

            NewMessage = new ChatMessage();
            NewMessage.PropertyChanged += NewMessage_PropertyChanged;
        }

        void NewMessage_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Message")
            {
                PostNewMessageCommand.RaiseCanExecuteChanged();
            }
        }


        public async void PostNewMessage()
        {
            ChatMessages.Add(NewMessage);

            _writer.WriteString(NewMessage.Message);

            await _writer.StoreAsync();

            CreateNewMessage();
        }

        public bool CanPostNewMessage()
        {
            //return NewMessage != null &&
            //    NewMessage.Message != null &&
            //    NewMessage.Message.Trim() != string.Empty;

            return true;
        }


        // Communications support
        private string _serverAddress;
        public string ServerAddress
        {
            get { return _serverAddress; }
            set { Set<string>(() => ServerAddress, ref _serverAddress, value); }
        }

        private string _connectionStatus;
        public string ConnectionStatus
        {
            get { return _connectionStatus; }
            set { Set<string>(() => ConnectionStatus, ref _connectionStatus, value); }
        }


        public RelayCommand ConnectCommand
        {
            get; private set;
        }

        private DataWriter _writer;
        private DataReader _reader;

        
        public async void Connect()
        {
            var hostName = new HostName(ServerAddress);

            _socket = new StreamSocket();

            await _socket.ConnectAsync(hostName, PortOrService);

            SetUpInputHandling();

            ConnectionStatus = "Connected to server at " + ServerAddress + ".";
        }

        public bool CanConnect()
        {
            return true;
        }




        public RelayCommand ListenCommand
        {
            get; private set;
        }

        public async void Listen()
        {
            _listener = new StreamSocketListener();
            _listener.ConnectionReceived += OnConnectionReceived;

            await _listener.BindServiceNameAsync(PortOrService);

            var hostNames = NetworkInformation.GetHostNames();

            ConnectionStatus = "Waiting for connection on: ";

            int i = 0;

            foreach (HostName name in hostNames)
            {
                if (i > 0)
                    ConnectionStatus += " and ";

                ConnectionStatus += name.DisplayName;
                i++;
            }

        }

        public bool CanListen()
        {
            return true;
        }


        void OnConnectionReceived(StreamSocketListener sender, StreamSocketListenerConnectionReceivedEventArgs args)
        {
            DispatcherHelper.CheckBeginInvokeOnUI(() =>
            {
                ConnectionStatus = "Connection received from " + 
                    args.Socket.Information.RemoteHostName.DisplayName;

                _socket = args.Socket;


                SetUpInputHandling();
            });
        }


        private void SetUpInputHandling()
        {
            _writer = new DataWriter(_socket.OutputStream);
            _reader = new DataReader(_socket.InputStream);

            var t = Task.Factory.StartNew(async () =>
            {
                _reader.InputStreamOptions = InputStreamOptions.Partial;

                while (true)
                {
                    var count = await _reader.LoadAsync(512);

                    var message = _reader.ReadString(count);

                    DispatcherHelper.CheckBeginInvokeOnUI(() =>
                    {
                        var chatMessage = new ChatMessage();
                        chatMessage.Message = message;

                        ChatMessages.Add(chatMessage);
                    });
                }
            });
        }


    }
}