﻿using GalaSoft.MvvmLight.Threading;
using SocketApp.Model;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Networking;
using Windows.Networking.Connectivity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;

namespace SocketApp.Services
{

    public class TcpStreamMessageService : IMessageService
    {
        private class PlayerConnection
        {
            public Player LocalPlayer { get; set; }
            public Player RemotePlayer { get; set; }
            public DataWriter Writer { get; set; }
            public DataReader Reader { get; set; }
            public StreamSocket Socket { get; set; }
        }

        public event EventHandler<ChatMessageReceivedEventArgs> ChatMessageReceived;
        public event EventHandler<ConnectionReceivedEventArgs> ConnectionReceived;
        public event EventHandler<PlayerJoinedEventArgs> PlayerJoined;
        public event EventHandler<PlayerExitedEventArgs> PlayerExited;

        private const string PortOrService = "5150";
        private const int MaxMessageSize = 1024;

        private StreamSocketListener _listener;
        private PlayerConnection _connection;

        public async void Connect(Player me, string remoteHostName)
        {
            var hostName = new HostName(remoteHostName);

            var pc = new PlayerConnection();

            pc = new PlayerConnection();
            pc.Socket = new StreamSocket();

            pc.LocalPlayer = me;
            pc.RemotePlayer = null;

            await pc.Socket.ConnectAsync(hostName, PortOrService);

            pc.Reader = new DataReader(pc.Socket.InputStream);
            pc.Writer = new DataWriter(pc.Socket.OutputStream);

            _connection = pc;

            ProcessIncomingMessages();

            SendPlayerJoinMessage();
        }


        private void CloseConnection()
        {
            if (_connection != null)
            {
                if (_connection.Writer != null)
                    _connection.Writer.Dispose();

                if (_connection.Reader != null)
                    _connection.Reader.Dispose();

                if (_connection.Socket != null)
                    _connection.Socket.Dispose();

                _connection.LocalPlayer = null;
                _connection.RemotePlayer = null;

                _connection = null;
            }
        }

        public void Disconnect()
        {
            SendPlayerLeaveMessage();

            CloseConnection();
        }

        public async void Listen(Player me)
        {
            _listener = new StreamSocketListener();
            _listener.ConnectionReceived += (s, e) =>
                {
                    var remoteHost = e.Socket.Information.RemoteHostName;

                    var pc = new PlayerConnection();

                    pc.Socket = e.Socket;
                    pc.Reader = new DataReader(pc.Socket.InputStream);
                    pc.Writer = new DataWriter(pc.Socket.OutputStream);
                    pc.LocalPlayer = me;
                    pc.RemotePlayer = null;

                    _connection = pc;

                    SendPlayerJoinMessage();

                    ProcessIncomingMessages();
                };

            await _listener.BindServiceNameAsync(PortOrService);
        }


        private void ProcessIncomingMessages()
        {
            var t = Task.Factory.StartNew(async () =>
            {
                _connection.Reader.InputStreamOptions = InputStreamOptions.Partial;

                while (true)
                {
                    var count = await _connection.Reader.LoadAsync(MaxMessageSize);

                    var messageType = (WireMessageType)_connection.Reader.ReadInt32();

                    switch (messageType)
                    {
                        case WireMessageType.PlayerLeave:
                            if (PlayerExited != null)
                                PlayerExited(this, new PlayerExitedEventArgs(_connection.RemotePlayer));
                            break;

                        case WireMessageType.PlayerJoin:
                            if (PlayerJoined != null)
                            {
                                var nameLength = _connection.Reader.ReadInt32();
                                var name = _connection.Reader.ReadString((uint)nameLength);

                                var remotePlayer = new Player();
                                remotePlayer.Name = name;

                                _connection.RemotePlayer = remotePlayer;

                                PlayerJoined(this, new PlayerJoinedEventArgs(remotePlayer));
                            }
                            break;

                        case WireMessageType.ChatMessage:
                            var msgLength = _connection.Reader.ReadInt32();
                            var text = _connection.Reader.ReadString((uint)msgLength);

                            var msg = new ChatMessage();
                            msg.Message = text;
                            msg.Player = _connection.RemotePlayer;

                            if (ChatMessageReceived != null)
                                ChatMessageReceived(this, new ChatMessageReceivedEventArgs(msg));
                            break;
                    }
                }
            });
        }


        private async void SendPlayerJoinMessage()
        {
            if (_connection.Writer != null)
            {
                string playerName = "(unknown)";
                if (_connection.LocalPlayer != null)
                    playerName = _connection.LocalPlayer.Name;

                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerJoin);
                _connection.Writer.WriteInt32((Int32)playerName.Length);
                _connection.Writer.WriteString(playerName);

                await _connection.Writer.StoreAsync();
            }
        }

        private async void SendPlayerLeaveMessage()
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerLeave);

                await _connection.Writer.StoreAsync();
            }
        }

        public async void SendChatMessage(ChatMessage message)
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.ChatMessage);
                _connection.Writer.WriteInt32(message.Message.Length);
                _connection.Writer.WriteString(message.Message);

                await _connection.Writer.StoreAsync();
            }
        }

        public IReadOnlyList<HostName> GetHostNames()
        {
            return NetworkInformation.GetHostNames();
        }
    }
}
