﻿using GalaSoft.MvvmLight.Threading;
using SocketApp.Model;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Networking;
using Windows.Networking.Connectivity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;

namespace SocketApp.Services
{

    public class TcpStreamMessageService : IMessageService
    {
        private class PlayerConnection
        {
            public Player LocalPlayer { get; set; }
            public Player RemotePlayer { get; set; }
            public DataWriter Writer { get; set; }
            public DataReader Reader { get; set; }
            public StreamSocket Socket { get; set; }
        }

        public event EventHandler<ChatMessageReceivedEventArgs> ChatMessageReceived;
        public event EventHandler<ConnectionReceivedEventArgs> ConnectionReceived;
        public event EventHandler<PlayerJoinedEventArgs> PlayerJoined;
        public event EventHandler<PlayerExitedEventArgs> PlayerExited;
        public event EventHandler<PlayerLocationUpdatedEventArgs> PlayerLocationUpdated;

        private const string PortOrService = "5150";
        private const int MaxMessageSize = 1024;

        private StreamSocketListener _listener;
        private PlayerConnection _connection;

        public async void Connect(Player me, string remoteHostName)
        {
            var hostName = new HostName(remoteHostName);

            var pc = new PlayerConnection();

            pc = new PlayerConnection();
            pc.Socket = new StreamSocket();

            pc.LocalPlayer = me;
            pc.RemotePlayer = null;

            await pc.Socket.ConnectAsync(hostName, PortOrService);

            pc.Reader = new DataReader(pc.Socket.InputStream);
            pc.Writer = new DataWriter(pc.Socket.OutputStream);

            _connection = pc;

            ProcessIncomingMessages();

            SendPlayerJoinMessage();
        }


        private void CloseConnection()
        {
            if (_connection != null)
            {
                if (_connection.Writer != null)
                    _connection.Writer.Dispose();

                if (_connection.Reader != null)
                    _connection.Reader.Dispose();

                if (_connection.Socket != null)
                    _connection.Socket.Dispose();

                _connection.LocalPlayer = null;
                _connection.RemotePlayer = null;

                _connection = null;
            }
        }

        public void Disconnect()
        {
            SendPlayerLeaveMessage();

            CloseConnection();
        }

        public async void Listen(Player me)
        {
            _listener = new StreamSocketListener();
            _listener.ConnectionReceived += (s, e) =>
                {
                    var remoteHost = e.Socket.Information.RemoteHostName;

                    var pc = new PlayerConnection();

                    pc.Socket = e.Socket;
                    pc.Reader = new DataReader(pc.Socket.InputStream);
                    pc.Writer = new DataWriter(pc.Socket.OutputStream);
                    pc.LocalPlayer = me;
                    pc.RemotePlayer = null;

                    _connection = pc;

                    SendPlayerJoinMessage();

                    ProcessIncomingMessages();
                };

            await _listener.BindServiceNameAsync(PortOrService);
        }


        private void ProcessIncomingMessages()
        {
            var t = Task.Factory.StartNew(async () =>
            {
                _connection.Reader.InputStreamOptions = InputStreamOptions.Partial;

                while (true)
                {
                    var count = await _connection.Reader.LoadAsync(MaxMessageSize);

                    var messageType = (WireMessageType)_connection.Reader.ReadInt32();

                    switch (messageType)
                    {
                        case WireMessageType.PlayerLeave:
                            if (PlayerExited != null)
                                PlayerExited(this, new PlayerExitedEventArgs(_connection.RemotePlayer));
                            break;

                        case WireMessageType.PlayerJoin:
                            if (PlayerJoined != null)
                            {
                                var nameLength = _connection.Reader.ReadInt32();
                                var name = _connection.Reader.ReadString((uint)nameLength);
                                var loc = ReadPlayerLocation(_connection.Reader);

                                var remotePlayer = new Player();
                                remotePlayer.Name = name;
                                remotePlayer.Location = loc;

                                _connection.RemotePlayer = remotePlayer;

                                PlayerJoined(this, new PlayerJoinedEventArgs(remotePlayer));
                            }
                            break;

                        case WireMessageType.ChatMessage:
                            if (ChatMessageReceived != null)
                            {
                                var msgLength = _connection.Reader.ReadInt32();
                                var text = _connection.Reader.ReadString((uint)msgLength);

                                var msg = new ChatMessage();
                                msg.Message = text;
                                msg.Player = _connection.RemotePlayer;

                                msg.Player.Location = ReadPlayerLocation(_connection.Reader);

                                ChatMessageReceived(this, new ChatMessageReceivedEventArgs(msg));
                            }
                            break;

                        case WireMessageType.PlayerLocation:
                            var player = _connection.RemotePlayer;
                            player.Location = ReadPlayerLocation(_connection.Reader);

                            Debug.WriteLine("Location received: " + player.Location.X + "," + player.Location.Y + " ^ " + player.Location.Angle);

                            if (PlayerLocationUpdated != null)
                                PlayerLocationUpdated(this, new PlayerLocationUpdatedEventArgs(player));
                            break;
                    
                    
                    }
                }
            });
        }



        private void WritePlayerLocation(DataWriter writer, PlayerLocation location)
        {
            writer.WriteDouble(location.X);
            writer.WriteDouble(location.Y);
            writer.WriteDouble(location.Angle);
        }

        private void WritePascalStyleString(string s)
        {
            if (s != null) s = s.Trim();

            if (s != null && s.Length > 0)
            {
                _connection.Writer.WriteInt32((Int32)s.Length);
                _connection.Writer.WriteString(s);
            }
            else
            {
                _connection.Writer.WriteInt32(0);
            }
        }

        private PlayerLocation ReadPlayerLocation(DataReader reader)
        {
            var loc = new PlayerLocation();

            loc.X = reader.ReadDouble();
            loc.Y = reader.ReadDouble();
            loc.Angle = reader.ReadDouble();

            return loc;
        }

        private async void SendPlayerJoinMessage()
        {
            if (_connection.Writer != null)
            {
                string playerName = "(unknown)";
                if (_connection.LocalPlayer != null)
                    playerName = _connection.LocalPlayer.Name;

                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerJoin);
                WritePascalStyleString(playerName);
                WritePlayerLocation(_connection.Writer, _connection.LocalPlayer.Location);

                await _connection.Writer.StoreAsync();
            }
        }

        private async void SendPlayerLeaveMessage()
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerLeave);

                await _connection.Writer.StoreAsync();
            }
        }

        public async void SendChatMessage(ChatMessage message)
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.ChatMessage);
                WritePascalStyleString(message.Message);
                WritePlayerLocation(_connection.Writer, _connection.LocalPlayer.Location);

                await _connection.Writer.StoreAsync();
            }
        }

        public async void SendLocationUpdate()
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerLocation);

                WritePlayerLocation(_connection.Writer, _connection.LocalPlayer.Location);
                await _connection.Writer.StoreAsync();
            }
        }



        public IReadOnlyList<HostName> GetHostNames()
        {
            return NetworkInformation.GetHostNames();
        }



    }
}
