﻿using GalaSoft.MvvmLight.Threading;
using SocketApp.Model;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Networking;
using Windows.Networking.Connectivity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;

namespace SocketApp.Services
{

    public class UdpMessageService : IMessageService
    {
        private class PlayerConnection
        {
            public Player LocalPlayer { get; set; }
            public Player RemotePlayer { get; set; }
            public DataWriter Writer { get; set; }
            //public DataReader Reader { get; set; }
            public DatagramSocket Socket { get; set; }
        }

        public event EventHandler<ChatMessageReceivedEventArgs> ChatMessageReceived;
        public event EventHandler<ConnectionReceivedEventArgs> ConnectionReceived;
        public event EventHandler<PlayerJoinedEventArgs> PlayerJoined;
        public event EventHandler<PlayerExitedEventArgs> PlayerExited;

        private const string PortOrService = "5150";
        private const int MaxMessageSize = 1024;

        private PlayerConnection _connection;

        public async void Connect(Player me, string remoteHostName)
        {
            _connection = new PlayerConnection();
            _connection.LocalPlayer = me;

            BindLocalSocket();

            ConnectToRemoteSocket(new HostName(remoteHostName), PortOrService);

            SendPlayerJoinMessage();
        }

        private async void BindLocalSocket()
        {
            _connection.Socket = new DatagramSocket();
            _connection.Socket.MessageReceived += OnSocketMessageReceived;
            await _connection.Socket.BindServiceNameAsync(PortOrService);
        }

        private async void ConnectToRemoteSocket(HostName hostName, string portOrService)
        {
            await _connection.Socket.ConnectAsync(hostName, portOrService);
            _connection.Writer = new DataWriter(_connection.Socket.OutputStream);

            SendPlayerJoinMessage();
        }

        async void OnSocketMessageReceived(DatagramSocket sender, 
                                        DatagramSocketMessageReceivedEventArgs args)
        {

            if (_connection.Writer == null)
                ConnectToRemoteSocket(args.RemoteAddress, args.RemotePort);

            var reader = args.GetDataReader();

            reader.InputStreamOptions = InputStreamOptions.Partial;

            //var count = await reader.LoadAsync(MaxMessageSize);

            var messageType = (WireMessageType)reader.ReadInt32();

            switch (messageType)
            {
                case WireMessageType.PlayerLeave:
                    if (PlayerExited != null)
                        PlayerExited(this, new PlayerExitedEventArgs(_connection.RemotePlayer));
                    break;

                case WireMessageType.PlayerJoin:
                    if (PlayerJoined != null)
                    {
                        var nameLength = reader.ReadInt32();
                        var name = reader.ReadString((uint)nameLength);

                        var remotePlayer = new Player();
                        remotePlayer.Name = name;

                        _connection.RemotePlayer = remotePlayer;

                        PlayerJoined(this, new PlayerJoinedEventArgs(remotePlayer));
                    }
                    break;

                case WireMessageType.ChatMessage:
                    var msgLength = reader.ReadInt32();
                    var text = reader.ReadString((uint)msgLength);

                    var msg = new ChatMessage();
                    msg.Message = text;
                    msg.Player = _connection.RemotePlayer;

                    if (ChatMessageReceived != null)
                        ChatMessageReceived(this, new ChatMessageReceivedEventArgs(msg));
                    break;
            }
        }


        private void CloseConnection()
        {
            if (_connection != null)
            {
                if (_connection.Writer != null)
                    _connection.Writer.Dispose();

                if (_connection.Socket != null)
                    _connection.Socket.Dispose();

                _connection.LocalPlayer = null;
                _connection.RemotePlayer = null;

                _connection = null;
            }
        }

        public void Disconnect()
        {
            SendPlayerLeaveMessage();

            CloseConnection();
        }

        public async void Listen(Player me)
        {
            _connection = new PlayerConnection();

            BindLocalSocket();
        }



        private async void SendPlayerJoinMessage()
        {
            if (_connection.Writer != null)
            {
                string playerName = "(unknown)";
                if (_connection.LocalPlayer != null)
                    playerName = _connection.LocalPlayer.Name;

                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerJoin);
                _connection.Writer.WriteInt32((Int32)playerName.Length);
                _connection.Writer.WriteString(playerName);

                await _connection.Writer.StoreAsync();
            }
        }

        private async void SendPlayerLeaveMessage()
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.PlayerLeave);

                await _connection.Writer.StoreAsync();
            }
        }

        public async void SendChatMessage(ChatMessage message)
        {
            if (_connection.Writer != null)
            {
                _connection.Writer.WriteInt32((Int32)WireMessageType.ChatMessage);
                _connection.Writer.WriteInt32(message.Message.Length);
                _connection.Writer.WriteString(message.Message);

                await _connection.Writer.StoreAsync();
            }
        }

        public IReadOnlyList<HostName> GetHostNames()
        {
            return NetworkInformation.GetHostNames();
        }
    }
}
