﻿using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Threading;
using SocketApp.Model;
using SocketApp.Services;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Threading.Tasks;
using Windows.Networking;
using Windows.Networking.Connectivity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using Windows.System.UserProfile;
using Windows.UI;
using Windows.UI.Xaml.Media;

namespace SocketApp.ViewModel
{
    public class MainViewModel : ViewModelBase
    {
        private IMessageService _messageService;

        private Random _random;

        public MainViewModel()
        {
            _random = new Random();

            _messageService = new TcpStreamMessageService();
            //_messageService = new UdpMessageService();

            _messageService.ChatMessageReceived += OnServiceChatMessageReceived;
            _messageService.ConnectionReceived += OnServiceConnectionReceived;
            _messageService.PlayerExited += OnServicePlayerExited;
            _messageService.PlayerJoined += OnServicePlayerJoined;

            ChatMessages = new ObservableCollection<ChatMessage>();
            ConnectionStatus = "Not connected.";
            ServerAddress = "pete-surface64";

            PostNewMessageCommand = new RelayCommand(
                () => PostNewMessage(),
                () => CanPostNewMessage());

            ListenCommand = new RelayCommand(
                () => Listen(),
                () => CanListen());

            ConnectCommand = new RelayCommand(
                () => Connect(),
                () => CanConnect());

            InitializePlayer();
            CreateNewMessage();


            //TestAddingPlayers();
        }




        private Player _player;
        public Player Player
        {
            get { return _player; }
            set { Set<Player>(() => Player, ref _player, value); }
        }

        public ObservableCollection<Player> Players { get; set; }

        private async void InitializePlayer()
        {
            Players = new ObservableCollection<Player>();

            Player = new Player();
            Player.Name = await UserInformation.GetDisplayNameAsync();
            Player.Color = GetNextPlayerColor();

            // set random starting position

            Player.Location.X = _random.Next(300) + 150;
            Player.Location.Y = _random.Next(200) + 75;
            Player.Location.Angle = _random.Next(359);

            //Debug.WriteLine(_player.Name + " x:" + _player.X + ", y:" + _player.Y + ", Angle:" + _player.Angle);

            Players.Add(_player);
        }


        private List<Color> _colors = new List<Color>()
            { Colors.Red, Colors.Green, Colors.Blue,
                Colors.Yellow, Colors.White, Colors.Violet,
                Colors.Azure, Colors.CornflowerBlue,
                Colors.Cyan, Colors.DeepPink, Colors.LightGray};

        private SolidColorBrush GetNextPlayerColor()
        {
            Color c;

            if (_colors.Count > 0)
            {
                // pick a color and then remove it from the list
                c = _colors[_random.Next(_colors.Count - 1)];
                _colors.Remove(c);
            }
            else
            {
                c = Colors.Wheat;
            }

            return new SolidColorBrush(c);
        }



        public void TestAddingPlayers()
        {
            for (int i = 0; i < 5; i++)
            {
                var p = new Player();
                p.Color = GetNextPlayerColor();

                p.Location.X = _random.Next(300) + 150;
                p.Location.Y = _random.Next(200) + 75;
                p.Location.Angle = _random.Next(359);
                p.Name = "Test Player " + i;

                Players.Add(p);
            }
        }

        public void TestPositionUpdate()
        {
            var t = Task.Factory.StartNew(async () =>
            {
                for (int i = 0; i < 100; i++)
                {
                    DispatcherHelper.CheckBeginInvokeOnUI(() =>
                        {
                            //double newX = Player.Location.X + 3;
                            //double newY = Player.Location.Y + 3;
                            //double newAngle = i * 5;

                            //Player.Location = new PlayerLocation() { X = newX, Y = newY, Angle = newAngle };

                            Player.Location.X += 3;
                            Player.Location.Y += 3;
                            Player.Location.Angle = i * 5;

                            _messageService.SendLocationUpdate();
                        });
                    
                    await Task.Delay(500);
                }
            });
        }



        public ObservableCollection<ChatMessage> ChatMessages { get; set; }

        private ChatMessage _newMessage;
        public ChatMessage NewMessage
        {
            get { return _newMessage; }
            set { Set<ChatMessage>(() => NewMessage, ref _newMessage, value); }
        }

        public RelayCommand PostNewMessageCommand
        {
            get; private set;
        }

        private void CreateNewMessage()
        {
            if (NewMessage != null)
                NewMessage.PropertyChanged -= NewMessage_PropertyChanged;

            NewMessage = new ChatMessage();
            NewMessage.Player = Player;
            NewMessage.PropertyChanged += NewMessage_PropertyChanged;
        }

        void NewMessage_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Message")
            {
                PostNewMessageCommand.RaiseCanExecuteChanged();
            }
        }


        public async void PostNewMessage()
        {
            ChatMessages.Add(NewMessage);
            _messageService.SendChatMessage(NewMessage);

            CreateNewMessage();
        }

        public bool CanPostNewMessage()
        {
            //return NewMessage != null &&
            //    NewMessage.Message != null &&
            //    NewMessage.Message.Trim() != string.Empty;

            return true;
        }


        // Communications support
        private string _serverAddress;
        public string ServerAddress
        {
            get { return _serverAddress; }
            set { Set<string>(() => ServerAddress, ref _serverAddress, value); }
        }

        private string _connectionStatus;
        public string ConnectionStatus
        {
            get { return _connectionStatus; }
            set { Set<string>(() => ConnectionStatus, ref _connectionStatus, value); }
        }


        public RelayCommand ConnectCommand
        {
            get; private set;
        }
       
        public void Connect()
        {
            _messageService.Connect(Player, ServerAddress);

            ConnectionStatus = "Connected to server at " + ServerAddress + ".";
        }

        public bool CanConnect()
        {
            return true;
        }




        public RelayCommand ListenCommand
        {
            get; private set;
        }

        public void Listen()
        {
            _messageService.Listen(Player);

            ConnectionStatus = "Waiting for connection on: ";

            int i = 0;

            foreach (HostName name in _messageService.GetHostNames())
            {
                if (i > 0)
                    ConnectionStatus += " and ";

                ConnectionStatus += name.DisplayName;
                i++;
            }
        }

        public bool CanListen()
        {
            return true;
        }


        void OnServicePlayerJoined(object sender, PlayerJoinedEventArgs e)
        {
            DispatcherHelper.CheckBeginInvokeOnUI(() =>
            {
                var p = e.Player;
                p.Color = GetNextPlayerColor();

                Players.Add(e.Player);

                ConnectionStatus = "Just joined: " +
                    e.Player.Name;
            });


            TestPositionUpdate();
        }

        void OnServicePlayerExited(object sender, PlayerExitedEventArgs e)
        {
            DispatcherHelper.CheckBeginInvokeOnUI(() =>
            {
                Players.Remove(e.Player);

                ConnectionStatus = "Just exited: " +
                    e.Player.Name;
            });
        }

        void OnServiceConnectionReceived(object sender, ConnectionReceivedEventArgs e)
        {
            DispatcherHelper.CheckBeginInvokeOnUI(() =>
            {
                ConnectionStatus = "Connection received from " +
                    e.HostName.DisplayName;
            });
        }

        void OnServiceChatMessageReceived(object sender, ChatMessageReceivedEventArgs e)
        {
            DispatcherHelper.CheckBeginInvokeOnUI(() =>
            {
                ChatMessages.Add(e.Message);
            });
        }
    }
}